///////////////////////////////////////////////////////////////////////////////
//
// File: DriverStandard.cpp
//
// For more information, please see: http://www.nektar.info
//
// The MIT License
//
// Copyright (c) 2006 Division of Applied Mathematics, Brown University (USA),
// Department of Aeronautics, Imperial College London (UK), and Scientific
// Computing and Imaging Institute, University of Utah (USA).
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included
// in all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
// THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
// DEALINGS IN THE SOFTWARE.
//
// Description: Driver class for the standard solver
//
///////////////////////////////////////////////////////////////////////////////

#include <iomanip>

#include <LibUtilities/BasicUtils/Timer.h>
#include <SolverUtils/DriverStandard.h>

namespace Nektar::SolverUtils
{
std::string DriverStandard::className =
    GetDriverFactory().RegisterCreatorFunction("Standard",
                                               DriverStandard::create);
std::string DriverStandard::driverLookupId =
    LibUtilities::SessionReader::RegisterEnumValue("Driver", "Standard", 0);

/**
 *
 */
DriverStandard::DriverStandard(
    const LibUtilities::SessionReaderSharedPtr pSession,
    const SpatialDomains::MeshGraphSharedPtr pGraph)
    : Driver(pSession, pGraph)
{
}

/**
 *
 */
void DriverStandard::v_InitObject(std::ostream &out)
{
    Driver::v_InitObject(out);
}

/**
 *
 */
void DriverStandard::v_Execute(std::ostream &out)

{
    Nektar::LibUtilities::Timer timer;
    NekDouble CPUtime;

    m_equ[0]->PrintSummary(out);

    timer.Start();

    m_equ[0]->DoInitialise();
    m_equ[0]->DoSolve();

    timer.Stop();

    m_equ[0]->Output();

    if (m_comm->GetRank() == 0)
    {
        CPUtime = timer.Elapsed().count();
        std::cout << "-------------------------------------------" << std::endl;
        std::cout << "Total Computation Time = " << CPUtime << "s" << std::endl;
        std::cout << "-------------------------------------------" << std::endl;
    }

    // Evaluate and output computation time and solution accuracy.
    // The specific format of the error output is essential for the
    // regression tests to work.

    // Evaluate L2 Error
    for (int i = 0; i < m_equ[0]->GetNvariables(); ++i)
    {
        Array<OneD, NekDouble> exactsoln(m_equ[0]->GetTotPoints(), 0.0);

        // Evaluate "ExactSolution" function, or zero array
        m_equ[0]->EvaluateExactSolution(i, exactsoln, m_equ[0]->GetTime());

        NekDouble vL2Error   = m_equ[0]->L2Error(i, exactsoln);
        NekDouble vLinfError = m_equ[0]->LinfError(i, exactsoln);

        if (m_comm->GetRank() == 0)
        {
            out << "L 2 error (variable " << m_equ[0]->GetVariable(i)
                << ") : " << vL2Error << std::endl;
            out << "L inf error (variable " << m_equ[0]->GetVariable(i)
                << ") : " << vLinfError << std::endl;
        }
    }
}
} // namespace Nektar::SolverUtils
